//
//  ControlTower.swift
//  ControlTower
//
//  Created by Pasan Premaratne on 3/22/16.
//  Copyright © 2016 Treehouse. All rights reserved.
//

typealias Knots = Int

// MARK: Protocols

protocol Flying {
    var descentSpeed: Knots { get }
}

protocol Landing {
    func requestLandingInstructions() -> LandingInstructions
}

protocol Airline: Flying, Landing {
    var type: AirlineType { get }
}

protocol AirlineType: Flying {}

struct LandingInstructions {
    let runway: ControlTower.Runway
    let terminal: ControlTower.Terminal
}

// MARK: Airline Type

enum DomesticAirlineType: AirlineType {
    case delta
    case american
    case united
}

enum InternationalAirlineType: AirlineType {
    case lufthansa
    case klm
    case airFrance
}

// MARK: Control Tower

final class ControlTower {
    
    enum Runway {
        case r22L
        case l31R
        case m52J
        case b19E
        
        static func runway(_ speed: Knots) -> Runway {
            switch speed {
            case 0..<91: return .r22L
            case 91...120: return .l31R
            case 121...140: return .m52J
            case 141...165: return .b19E
            default: return .b19E
            }
        }
    }
    
    enum Terminal {
        case a(Int?)
        case b(Int?)
        case c(Int?)
        case international(Int?)
        case `private`(Int?)
        
        static func terminal(_ airline: Airline) -> Terminal {
            switch airline.type {
            case is DomesticAirlineType:
                let domesticAirline = airline.type as! DomesticAirlineType
                switch domesticAirline {
                case .delta:
                    let gate = GateManager.sharedInstance.gateFor(.a(nil))
                    return .a(gate)
                case .american:
                    let gate = GateManager.sharedInstance.gateFor(.b(nil))
                    return .b(gate)
                case .united:
                    let gate = GateManager.sharedInstance.gateFor(.c(nil))
                    return .c(gate)
                }
            case is InternationalAirlineType:
                let gate = GateManager.sharedInstance.gateFor(.international(nil))
                return .international(gate)
            default:
                let gate = GateManager.sharedInstance.gateFor(.private(nil))
                return .private(gate)
            }
        }
    }
    
    class GateManager {
        
        static let sharedInstance = GateManager()
        
        fileprivate init() {}
        
        var gatesForTerminalA: [String: [Int]] = ["occupied": [1,2,3,4,5,6,7,8], "empty": [9,10,11,12]]
        var gatesForTerminalB: [String: [Int]] = ["occupied": [1], "empty": [2,3,4,5,6,7,8]]
        var gatesForTerminalC: [String: [Int]] = ["occupied": [1,2,3,4], "empty": [5,6,7,8,9,10]]
        var gatesForInternationalTerminal: [String: [Int]] = ["occupied": [1,2,3], "empty": [4,5,6]]
        var gatesForPrivateHangar: [String: [Int]] = ["occupied": [1], "empty": [2,3]]
        
        func update(_ gates: inout [String: [Int]], gate: Int) {
            if var occupiedGates = gates["occupied"] {
                occupiedGates.append(gate)
                gates.updateValue(occupiedGates, forKey: "occupied")
            }
            
            if var emptyGates = gates["empty"], let index = emptyGates.index(of: gate) {
                emptyGates.remove(at: index)
                gates.updateValue(emptyGates, forKey: "empty")
            }
        }
        
        func emptyGate(_ gates: inout [String: [Int]]) -> Int? {
            guard let gate = gates["empty"]?.first else {
                return nil
            }
            
            update(&gates, gate: gate)
            return gate
        }
        
        func gateFor(_ terminal: Terminal) -> Int? {
            switch terminal {
            case .a:
                return emptyGate(&gatesForTerminalA)
            case .b:
                return emptyGate(&gatesForTerminalB)
            case .c:
                return emptyGate(&gatesForTerminalC)
            case .international:
                return emptyGate(&gatesForInternationalTerminal)
            case .private:
                return emptyGate(&gatesForPrivateHangar)
            }
        }
    }
    
    func land(_ airline: Airline) -> LandingInstructions {
        let runway = Runway.runway(airline.descentSpeed)
        let terminal = Terminal.terminal(airline)
        return LandingInstructions(runway: runway, terminal: terminal)
    }
}

extension Airline {
    
    var descentSpeed: Knots {
        return type.descentSpeed
    }
    
    func requestLandingInstructions() -> LandingInstructions {
        return ControlTower().land(self)
    }
}

extension DomesticAirlineType {
    var descentSpeed: Knots {
        return 100
    }
}

extension InternationalAirlineType {
    var descentSpeed: Knots {
        return 130
    }
}

struct Flight: Airline {
    let type: AirlineType
}




















